TEXT = """
Overhead the albatross hangs motionless upon the air
And deep beneath the rolling waves in labyrinths of coral caves
The echo of a distant tide
Comes willowing across the sand
And everything is green and submarine

And no one showed us to the land
And no one knows the wheres or whys
But something stirs and something tries
And starts to climb towards the light

Strangers passing in the street
By chance two separate glances meet
And I am you and what I see is me
And do I take you by the hand
And lead you through the land
And help me understand the best I can
And no one calls us to move on
And no one forces down our eyes
No one speaks
And no one tries
No one flies around the sun

...

Echoes - Pink Floyd
"""

# Compression parameters:
#    window_sz2 - 11
#    lookahead_sz2 - 4
COMPRESSED = (
    b'\x85S\xee\xd6[\x95\xa2\xcba\xb2H.\x96\x8b,\x82\xc3l\xb1Xn\x97+}\xce\xe7 '
    b'\xb4Xm\xd6{\x9c\x82\xdbo\xbaZm\xf6\xebe\x94\x04\n\xebp\xb7\xdb\x80\x8dm7'
    b'(U\x06\xddd\x90Y,\xb6[\x84\x82\xc5e\xb7Yl7K@\x0c%\xca\xdfl\xb6Zm\xd6y'
    b'\x05\xde\xc3v\xb2\xdc\xe4\x16\x9bt\x82\xd9a\xb1^nV\x9bu\xd2\xd1s\x90['
    b'\xec\xd2\x0b\x1d\xbe\xe5a\xb6H,`8p\xaa\xa0\x17\x15\x96\xc7h\xb7\x80\xb9'
    b'\xd8d\x16KM\xce\xe9a\xb7]$\x17KM\x92\xcb\n\xa1\xdb\xed\xa0zW{M\xb2\xd9o'
    b'\xbb\x82i\xd8la\x1e\x80\xc0w;\r\xba\xc8\x0fIe\x0b2\xbc\xdd-\x00<v\x9b'
    b'\x9c\x82\xcfr\xb2\xd9m\xd2\x00\x1b)\x05\xce\xebb\xb6\xd8A\xc4\xac\xb0'
    b'\xa0&Kp/\x95\xb8\r\x8a\xd1o\xbb\xd9l\x92\x0b\xa8"\x15\xbc\x11\xd2\xd8'
    b'\x08\xee\x04\rk\xb7[\xee\xe0\xc2\xb7{E\x96\xe4\x0f\xa5o\xb9\x00I^np\xaa'
    b'\x15\xd6\xe9 \xb9\x84y\x03R\xdc\xee\x96\x9b\x95\xcc\x19\xd4\x04\xe2\xe9r'
    b'\xb4\x86q\x02Y\x86\x11\\\xae\x80\xc4\x96;e\xa6\xdbb\x00\x82\xbb\xd8nV@V'
    b'[e\xa6\xcfh\xbaB\xa1U;\xa5\xc8UJ\xca\x08\x85p\xb0\xdc\xeeaz\x82\xa4\xb7;'
    b'\xa0a\x15\xd2\x15B\xbc\xc8,b\xeeV;,\x82\xe9w\xb7\xc8.v[\x85\x86\xe5a'
    b'\xbaYd\x16p\xaf+\x18J\x95\xb4\x08\xcc\x1a\x8e\x93 \xb0\xdbd\x17\x9b}\xd4"'
    b'\xd2\xefh\xb0\xdd\x00D\xaev[(\x85\x9d\xb6\xca3ko\x01"\xbaXmvP&K\x15\xe4'
    b'\x19P\x16\x88=\x96\xd8=\xc6\x03\x07t\xb4\\\xad\xf7[8\xde\xa8\x8aE\xa2'
    b'\xcbl\xb8\x029H.\xb6\xeb U\x10\xcc\x11\r-\x8a\xcbs\x06s\xb1\xd8m\xc2uv;'
    b'\r\xb2\xd8Cd\x1e\xe9m\xb7\xdd\xac\xa0(@=vk}\xc8.N\xc9o\xbb\xdb\xa4\x16'
    b'\xfb\xad\xcaAe\xbc\x89iS\x86Y\xae\x16[\r\xacN\xd4\x0bXS\x14\x07\x9a\xcdl'
    b'\x00\xc2\x90XB\x80\x82\x01\xaew[t*\x15.\x97K\xa1P\xaa)\x1c\x90\r\x94'
    b'\xb6AP\xb4\xdb\xadr\n5\xb2\xdfy\xb2B\x80'
)
